# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import pytest
import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tripleo import undercloud
from tobiko.tripleo import nova as tripleo_nova


@pytest.mark.minimal
class NetworkTest(testtools.TestCase):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.CirrosPeerServerStackFixture)

    def test_stack_create_complete(self):
        self.stack.key_pair_stack.wait_for_create_complete()
        self.stack.network_stack.wait_for_create_complete()
        self.stack.peer_stack.wait_for_create_complete()
        self.stack.wait_for_create_complete()

    def test_ssh(self):
        """Test TCP connectivity to SSH server from VM to VM"""
        hostname = sh.ssh_hostname(ssh_client=self.stack.ssh_client)
        self.assertEqual(self.stack.server_name.lower(), hostname)

    def test_ping(self):
        """Test ICMP connectivity to from VM to VM"""
        ping.assert_reachable_hosts(
            [self.stack.ip_address],
            ssh_client=self.stack.peer_stack.ssh_client)

    # --- test l3_ha extension ------------------------------------------------

    @neutron.skip_unless_is_ovs()
    @neutron.skip_if_missing_networking_extensions('l3-ha')
    def test_l3_ha(self):
        """Test l3-ha network attribute"""
        gateway = self.stack.network_stack.gateway_details
        self.assertEqual(self.stack.network_stack.ha,
                         gateway['ha'])


@pytest.mark.background
@undercloud.skip_if_missing_undercloud
class BackgroundProcessTest(NetworkTest):

    def test_check_background_vm_ping(self):
        """ Tests that are designed to run in the background ,
            then collect results.
            Logic: checks if process exists, if so stop the process,
            then execute some check logic i.e. a check function.
            if the process by name isn't running,
            start a separate process i.e a background function"""
        tripleo_nova.check_or_start_background_vm_ping()


@pytest.mark.migrate_server
class SameHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosSameHostServerStackFixture)

    def test_same_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'same_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertEqual(nova.get_server_hypervisor(sender),
                         nova.get_server_hypervisor(receiver))


@pytest.mark.migrate_server
@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class DifferentHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosDifferentHostServerStackFixture)

    def test_different_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'different_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertNotEqual(nova.get_server_hypervisor(sender),
                            nova.get_server_hypervisor(receiver))


# --- l3-ha extension VM to VM scenario ---------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haNetworkTest(NetworkTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.L3haPeerServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haSameHostNetworkTest(SameHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.L3haSameHostServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class L3haDifferentHostNetworkTest(DifferentHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.L3haDifferentHostServerStackFixture)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack import keystone
from tobiko import tripleo


def load_shiftstack_rcfile() -> typing.Dict[str, str]:
    conf = tobiko.tobiko_config().shiftstack
    return tripleo.fetch_os_env(*conf.rcfile)


class ShiftstackKeystoneCredentialsFixture(
        tripleo.UndercloudCloudsFileKeystoneCredentialsFixture):

    @staticmethod
    def _get_default_cloud_name() -> typing.Optional[str]:
        return tobiko.tobiko_config().shiftstack.cloud_name

    def _get_environ(self) -> typing.Dict[str, str]:
        return load_shiftstack_rcfile()


def shiftstack_keystone_session() -> keystone.KeystoneSession:
    credentials = shiftstack_keystone_credentials()
    return keystone.get_keystone_session(credentials=credentials)


def shiftstack_keystone_credentials() -> keystone.KeystoneCredentialsFixture:
    return tobiko.get_fixture(ShiftstackKeystoneCredentialsFixture)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

import netaddr

import tobiko
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _network


SubnetType = typing.Dict[str, typing.Any]
SubnetIdType = typing.Union[str, SubnetType]


def get_subnet_id(subnet: SubnetIdType) -> str:
    if isinstance(subnet, str):
        return subnet
    else:
        return subnet['id']


def get_subnet(subnet: SubnetIdType,
               client: _client.NeutronClientType = None,
               **params) -> SubnetType:
    subnet_id = get_subnet_id(subnet)
    try:
        return _client.neutron_client(client).show_subnet(
            subnet_id, **params)['subnet']
    except _client.NotFound as ex:
        raise NoSuchSubnet(id=subnet_id) from ex


def create_subnet(client: _client.NeutronClientType = None,
                  network: _network.NetworkIdType = None,
                  add_cleanup=True,
                  ip_version=4,
                  **params) -> SubnetType:
    if 'network_id' not in params:
        if network is None:
            from tobiko.openstack import stacks
            network_id = tobiko.setup_fixture(
                stacks.NetworkStackFixture).network_id
        else:
            network_id = _network.get_network_id(network)
        params['network_id'] = network_id
    params['ip_version'] = ip_version
    subnet = _client.neutron_client(client).create_subnet(
        body={'subnet': params})['subnet']
    if add_cleanup:
        tobiko.add_cleanup(cleanup_subnet, subnet=subnet, client=client)
    return subnet


def cleanup_subnet(subnet: SubnetIdType,
                   client: _client.NeutronClientType = None):
    try:
        delete_subnet(subnet=subnet, client=client)
    except NoSuchSubnet:
        pass


def delete_subnet(subnet: SubnetIdType,
                  client: _client.NeutronClientType = None):
    subnet_id = get_subnet_id(subnet)
    try:
        _client.neutron_client(client).delete_subnet(subnet_id)
    except _client.NotFound as ex:
        raise NoSuchSubnet(id=subnet_id) from ex


def list_subnets(client: _client.NeutronClientType = None,
                 ip_version: int = None,
                 network: _network.NetworkIdType = None,
                 **params) -> tobiko.Selection[SubnetType]:
    if ip_version is not None:
        params['ip_version'] = ip_version
    if network is not None:
        params.setdefault('network_id', _network.get_network_id(network))
    subnets = _client.neutron_client(client).list_subnets(**params)
    if isinstance(subnets, abc.Mapping):
        subnets = subnets['subnets']
    return tobiko.select(subnets)


def find_subnet(client: _client.NeutronClientType = None,
                unique=False,
                default: SubnetType = None,
                **params) -> SubnetType:
    """Look for a subnet matching some property values"""
    subnets = list_subnets(client=client, **params)
    if default is None or subnets:
        if unique:
            return subnets.unique
        else:
            return subnets.first
    else:
        return default


def update_subnet(subnet: SubnetIdType,
                  client: _client.NeutronClientType = None,
                  gateway_ip: typing.Union[str, netaddr.IPAddress] = None,
                  **params) -> SubnetType:
    subnet_id = get_subnet_id(subnet)
    if gateway_ip is not None:
        params['gateway_ip'] = str(gateway_ip)
    try:
        return _client.neutron_client(client).update_subnet(
                subnet_id, body={'subnet': params})['subnet']
    except _client.NotFound as ex:
        raise NoSuchSubnet(id=subnet_id) from ex


def ensure_subnet_gateway(subnet: SubnetIdType,
                          client: _client.NeutronClientType = None) \
        -> SubnetType:
    """Make sure given subnet has a gateway IP"""
    if isinstance(subnet, str):
        subnet = get_subnet(subnet, client=client)
    if not subnet.get('gateway_ip'):
        cidr = netaddr.IPNetwork(subnet['cidr'])
        gateway_ip = netaddr.IPAddress(cidr.first + 1)
        subnet = update_subnet(subnet=subnet,
                               client=client,
                               gateway_ip=gateway_ip)
    return subnet


class NoSuchSubnet(tobiko.ObjectNotFound):
    message = "No such subnet found for {id!r}"



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from neutronclient.v2_0 import client as neutronclient

import tobiko
from tobiko.openstack import _client


NeutronClientException = neutronclient.exceptions.NeutronClientException
NotFound = neutronclient.exceptions.NotFound
ServiceUnavailable = neutronclient.exceptions.ServiceUnavailable


class NeutronClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return neutronclient.Client(session=session)


class NeutronClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return NeutronClientFixture(session=session)


CLIENTS = NeutronClientManager()


NeutronClient = neutronclient.Client
NeutronClientType = typing.Union[neutronclient.Client,
                                 NeutronClientFixture]


def neutron_client(obj: NeutronClientType = None) \
        -> neutronclient.Client:
    if obj is None:
        return get_neutron_client()

    if isinstance(obj, neutronclient.Client):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, NeutronClientFixture):
        return fixture.client

    message = "Object {!r} is not a NeutronClientFixture".format(obj)
    raise TypeError(message)


class HasNeutronClientFixture(tobiko.SharedFixture):

    client: typing.Optional[neutronclient.Client] = None

    def __init__(self, obj: NeutronClientType = None):
        super(HasNeutronClientFixture, self).__init__()
        self._obj = obj

    def setup_fixture(self):
        self.client = neutron_client(self._obj)


def get_neutron_client(session=None, shared=True, init_client=None,
                       manager=None) -> neutronclient.Client:
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client



#!/usr/bin/env python3
# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import site
import sys


TOP_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))

if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa
from tools import get_version  # noqa

LOG = common.get_logger(__name__)

TOX_VERSION = os.environ.get('TOX_VERSION') or '>=4.13'

TOX_CONSTRAINTS = (os.environ.get('TOX_CONSTRAINTS') or
                   f"-c{TOP_DIR}/upper-constraints.txt")


def main():
    common.setup_logging()
    install_tox()
    install_bindeps()
    install_tobiko()


def install_tox(version=TOX_VERSION):
    LOG.info(f"Installing Tox... (version: {version})")
    pip_install(f"'tox{version}'")


def install_bindeps():
    LOG.info(f"Installing Tobiko binary dependencies...")
    common.execute(os.path.join(TOP_DIR, 'tools', 'install-bindeps.sh'),
                   capture_stdout=False)


def install_tobiko():
    version = get_version.get_version()
    LOG.info(f"Installing Tobiko version {version}...")
    pip_install(f"-e '{TOP_DIR}'")


def pip_install(args):
    LOG.debug(f"Installing packages: {args}...")
    common.execute_python(f"-m pip install {TOX_CONSTRAINTS} {args}",
                          capture_stdout=False)


def pip_unisntall(args):
    LOG.debug(f"Uninstalling packages: {args}...")
    common.execute_python(f"-m pip uninstall -y {args}",
                          capture_stdout=False)


if __name__ == '__main__':
    main()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from glanceclient.v2 import client as glanceclient
from glanceclient import exc
from oslo_log import log

import tobiko
from tobiko.openstack import _client

LOG = log.getLogger(__name__)


class GlanceClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return glanceclient.Client(session=session)


class GlanceClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return GlanceClientFixture(session=session)


CLIENTS = GlanceClientManager()


def glance_client(obj=None):
    obj = obj or default_glance_client()
    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client
    return tobiko.check_valid_type(obj, glanceclient.Client)


def default_glance_client():
    return get_glance_client()


def get_glance_client(session=None, shared=True, init_client=None,
                      manager=None):
    manager = manager or CLIENTS
    fixture = manager.get_client(session=session, shared=shared,
                                 init_client=init_client)
    return glance_client(fixture)


def create_image(client=None, **params):
    """Look for the unique network matching some property values"""
    return glance_client(client).images.create(**params)


def delete_image(image_id, client=None, **params):
    try:
        glance_client(client).images.delete(image_id, **params)
    except exc.HTTPNotFound:
        return False
    else:
        return True


_RAISE_ERROR = object()


def get_image(image_id, client=None, default=_RAISE_ERROR):
    try:
        return glance_client(client).images.get(image_id=image_id)
    except exc.HTTPNotFound:
        if default is _RAISE_ERROR:
            raise
        else:
            return default


def find_image(image=None, client=None, unique=False, default=_RAISE_ERROR,
               **params):
    """Look for an image matching some property values"""
    client = glance_client(client)
    if image:
        try:
            return get_image(image_id=image, default=_RAISE_ERROR)
        except exc.HTTPNotFound:
            params.setdefault('name', image)

    limit = unique and 2 or 1
    images = list_images(client=client, limit=limit, **params)
    if default is _RAISE_ERROR or images:
        if unique:
            return images.unique
        else:
            return images.first
    else:
        return default


def list_images(client=None, limit=None, **filters):
    images = glance_client(client).images.list(limit=limit, filters=filters)
    return tobiko.select(images)


def upload_image(image_id, image_data, client=None, **params):
    """Look for the unique network matching some property values"""
    return glance_client(client).images.upload(
        image_id=image_id, image_data=image_data, **params)


class HasGlanceClientMixin(object):

    glance_client = None

    def get_image(self, image_id, **params):
        return get_image(image_id=image_id,
                         client=self.glance_client,
                         **params)

    def find_image(self, **params):
        return find_image(client=self.glance_client, **params)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.keystone import _client
from tobiko.openstack.keystone import _clouds_file
from tobiko.openstack.keystone import _credentials
from tobiko.openstack.keystone import _resource
from tobiko.openstack.keystone import _services
from tobiko.openstack.keystone import _session

KeystoneClient = _client.KeystoneClient
KeystoneClientFixture = _client.KeystoneClientFixture
keystone_client = _client.keystone_client
get_keystone_client = _client.get_keystone_client
find_service = _client.find_service
find_endpoint = _client.find_endpoint
find_service_endpoint = _client.find_service_endpoint
list_endpoints = _client.list_endpoints
list_services = _client.list_services

CloudsFileKeystoneCredentialsFixture = (
    _clouds_file.CloudsFileKeystoneCredentialsFixture)

default_keystone_credentials = _credentials.default_keystone_credentials
has_keystone_credentials = _credentials.has_keystone_credentials
keystone_credentials = _credentials.keystone_credentials
register_default_keystone_credentials = (
    _credentials.register_default_keystone_credentials)
skip_unless_has_keystone_credentials = (
    _credentials.skip_unless_has_keystone_credentials)
ConfigKeystoneCredentialsFixture = (
    _credentials.ConfigKeystoneCredentialsFixture)
DelegateKeystoneCredentialsFixture = (
    _credentials.DelegateKeystoneCredentialsFixture)
KeystoneCredentials = _credentials.KeystoneCredentials
KeystoneCredentialsFixture = _credentials.KeystoneCredentialsFixture
KeystoneCredentialsType = _credentials.KeystoneCredentialsType
NoSuchKeystoneCredentials = _credentials.NoSuchKeystoneCredentials
EnvironKeystoneCredentialsFixture = \
    _credentials.EnvironKeystoneCredentialsFixture
InvalidKeystoneCredentials = _credentials.InvalidKeystoneCredentials

get_keystone_resource_id = _resource.get_keystone_resource_id
get_project_id = _resource.get_project_id
get_user_id = _resource.get_user_id
KeystoneResourceType = _resource.KeystoneResourceType
ProjectType = _resource.ProjectType
UserType = _resource.UserType

has_service = _services.has_service
is_service_missing = _services.is_service_missing
skip_if_missing_service = _services.skip_if_missing_service

keystone_session = _session.keystone_session
KeystoneSession = _session.KeystoneSession
KeystoneSessionType = _session.KeystoneSessionType
KeystoneSessionFixture = _session.KeystoneSessionFixture
KeystoneSessionManager = _session.KeystoneSessionManager
get_keystone_endpoint = _session.get_keystone_endpoint
get_keystone_session = _session.get_keystone_session
get_keystone_token = _session.get_keystone_token



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import sys
import typing
import unittest

from oslo_log import log

import tobiko
from tobiko.run import _config
from tobiko.run import _discover


LOG = log.getLogger(__name__)


def run_tests(test_path: typing.Union[str, typing.Iterable[str]],
              test_filename: str = None,
              python_path: typing.Iterable[str] = None,
              config: _config.RunConfigFixture = None,
              result: unittest.TestResult = None,
              check=True) -> unittest.TestResult:
    test_ids = _discover.find_test_ids(test_path=test_path,
                                       test_filename=test_filename,
                                       python_path=python_path,
                                       config=config)
    return run_test_ids(test_ids=test_ids, result=result, check=check)


def run_test_ids(test_ids: typing.List[str],
                 result: unittest.TestResult = None,
                 check=True) \
        -> unittest.TestResult:
    test_classes: typing.Dict[str, typing.List[str]] = \
        collections.defaultdict(list)

    # regroup test ids my test class keeping test names order
    test_ids = list(test_ids)
    for test_id in test_ids:
        test_class_id, test_name = test_id.rsplit('.', 1)
        test_classes[test_class_id].append(test_name)

    # add test cases to the suite ordered by class name
    suite = unittest.TestSuite()
    for test_class_id, test_names in sorted(test_classes.items()):
        test_class = tobiko.load_object(test_class_id)
        for test_name in test_names:
            test = test_class(test_name)
            suite.addTest(test)

    LOG.info(f'Run {len(test_ids)} test(s)')
    result = tobiko.run_test(case=suite, result=result, check=check)

    LOG.info(f'{result.testsRun} test(s) run')
    return result


class RunTestCasesFailed(tobiko.TobikoException):
    message = ('Test case execution failed:\n'
               '{errors}\n'
               '{failures}\n')


def main(test_path: typing.Iterable[str] = None,
         test_filename: str = None,
         python_path: typing.Iterable[str] = None):
    if test_path is None:
        test_path = sys.argv[1:]

    result = run_tests(test_path=test_path,
                       test_filename=test_filename,
                       python_path=python_path)

    for case, exc_info in result.errors:
        LOG.exception(f"Test case error: {case.id()}",
                      exc_info=exc_info)

    for case, exc_info in result.errors:
        LOG.exception(f"Test case failure: {case.id()}",
                      exc_info=exc_info)

    for case, reason in result.skipped:
        LOG.info(f"Test case skipped: {case.id()} ({reason})")

    LOG.info(f"{result.testsRun} test case(s) executed:\n"
             f"  errors:   {len(result.errors)}"
             f"  failures: {len(result.failures)}"
             f"  skipped:  {len(result.skipped)}")
    if result.errors or result.failures:
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == '__main__':
    main()



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.rhosp import _version_utils
from tobiko.shell import ssh

LOG = log.getLogger(__name__)


def get_ip_to_nodes_dict(group, openstack_nodes=None):
    if not openstack_nodes:
        openstack_nodes = topology.list_openstack_nodes(group=group)
    ip_to_nodes_dict = {str(node.public_ip): node.name for node in
                        openstack_nodes}
    return ip_to_nodes_dict


def ip_to_hostname(oc_ip, group=None):
    ip_to_nodes_dict = get_ip_to_nodes_dict(group)
    oc_ipv6 = oc_ip.replace(".", ":")
    if netaddr.valid_ipv4(oc_ip) or netaddr.valid_ipv6(oc_ip):
        return ip_to_nodes_dict[oc_ip]
    elif netaddr.valid_ipv6(oc_ipv6):
        LOG.debug("The provided string was a modified IPv6 address: %s",
                  oc_ip)
        return ip_to_nodes_dict[oc_ipv6]
    else:
        tobiko.fail("wrong IP value provided %s" % oc_ip)


class RhospTopology(topology.OpenStackTopology):

    """Base topology for Red Hat OpenStack deployments.

    This is base topology which represents common parts between Tripleo and
    Podified deployments.
    """

    has_containers = True
    container_runtime_cmd = 'podman'

    @property
    def ignore_containers_list(self):
        return None


class RhospNode(topology.OpenStackTopologyNode):

    """Base RHOSP Node

    This class represents common parts between Overcloud nodes and EDPM nodes
    in Red Hat OpenStack deployments.
    """

    def __init__(self,
                 topology: topology.OpenStackTopology,
                 name: str,
                 ssh_client: ssh.SSHClientFixture,
                 addresses: typing.Iterable[netaddr.IPAddress],
                 hostname: str,
                 rhosp_version: tobiko.Version = None):
        # pylint: disable=redefined-outer-name
        super().__init__(topology=topology,
                         name=name,
                         ssh_client=ssh_client,
                         addresses=addresses,
                         hostname=hostname)
        self._rhosp_version = rhosp_version

    @property
    def rhosp_version(self) -> tobiko.Version:
        if self._rhosp_version is None:
            self._rhosp_version = self._get_rhosp_version()
        return self._rhosp_version

    def _get_rhosp_version(self) -> tobiko.Version:
        return _version_utils.get_rhosp_version(connection=self.connection)

    def power_on_node(self):
        pass

    def power_off_node(self):
        pass

    def reboot_node(self, reactivate_servers=True):
        """Reboot node

        This method reboots a node and may start every Nova
        server which is not in SHUTOFF status before restarting.

        :param reactivate_servers: whether or not to re-start the servers which
            are hosted on the compute node after the reboot
        """

        running_servers: typing.List[nova.NovaServer] = []
        if reactivate_servers:
            running_servers = self.list_running_servers()
            LOG.debug(f'Servers to restart after reboot: {running_servers}')

        self.power_off_node()
        self.power_on_node()

        if running_servers:
            LOG.info(f'Restart servers after rebooting compute node '
                     f'{self.name}...')
            for server in running_servers:
                nova.wait_for_server_status(server=server.id,
                                            status='SHUTOFF')
                LOG.debug(f'Re-activate server {server.name} with ID '
                          f'{server.id}')
                nova.activate_server(server=server)
                LOG.debug(f'Server {server.name} with ID {server.id} has '
                          f'been reactivated')

    def list_running_servers(self) -> typing.List[nova.NovaServer]:
        running_servers = list()
        for server in nova.list_servers():
            if server.status != 'SHUTOFF':
                hypervisor_name = nova.get_server_hypervisor(server,
                                                             short=True)
                if self.name == hypervisor_name:
                    running_servers.append(server)
        return running_servers



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko.shell import ip
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import topology


LOG = log.getLogger(__name__)


@pytest.mark.minimal
class OpenstackNodesTest(testtools.TestCase):

    topology = tobiko.required_fixture(
        topology.get_default_openstack_topology_class())

    def test_public_ips(self):
        ips = dict()
        for node in self.topology.nodes:
            ping.ping(node.public_ip).assert_replied()
            other = ips.setdefault(node.public_ip, node)
            if node is not other:
                tobiko.fail(f"Nodes {node.name} and {other.name} have the "
                            f"same IP: {node.public_ip}")

    def test_hostnames(self):
        hostnames = dict()
        for node in self.topology.nodes:
            if node.ssh_client is None:
                LOG.debug(f'Node {node.hostname} has no ssh_client')
                continue
            hostname = sh.get_hostname(ssh_client=node.ssh_client)
            self.assertTrue(hostname.startswith(node.name))
            other = hostnames.setdefault(hostname, node)
            if node is not other:
                tobiko.fail(f"Nodes {node.name} and {other.name} have the "
                            f"same hostname: {hostname}")

    def test_network_namespaces(self):
        for node in self.topology.nodes:
            namespaces_ips = {}
            namespaces = ip.list_network_namespaces(ssh_client=node.ssh_client)
            for namespace in namespaces:
                ips = ip.list_ip_addresses(ssh_client=node.ssh_client,
                                           network_namespace=namespace)
                other_ips = namespaces_ips.setdefault(namespace, ips)
                if ips is not other_ips:
                    tobiko.fail(f"Duplicate network namespace {namespace} in "
                                f"node {node.name}: {other_ips}, {ips}")



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from manilaclient.v2 import client as manilaclient
from manilaclient import exceptions
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import _client
from tobiko.openstack.manila import _exceptions

LOG = log.getLogger(__name__)
CONF = config.CONF


class ManilaClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        cacert = (session.cert or CONF.tobiko.tripleo.undercloud_cacert_file
                  if 'https://' in session.auth.auth_url
                  else None)
        return manilaclient.Client(session=session, cacert=cacert)


class ManilaClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return ManilaClientFixture(session=session)


CLIENTS = ManilaClientManager()


@keystone.skip_if_missing_service(name='manila')
def manila_client(obj=None):
    obj = obj or default_manila_client()
    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client
    return tobiko.check_valid_type(obj, manilaclient.Client)


def default_manila_client():
    return get_manila_client()


def get_manila_client(session=None, shared=True, init_client=None,
                      manager=None):
    manager = manager or CLIENTS
    fixture = manager.get_client(session=session, shared=shared,
                                 init_client=init_client)
    return manila_client(fixture)


def create_share(share_protocol=None, size=None, client=None, **kwargs):
    share_protocol = share_protocol or CONF.tobiko.manila.share_protocol
    share_size = size or CONF.tobiko.manila.size
    return manila_client(client).shares.create(
        share_proto=share_protocol, size=share_size, return_raw=True,
        **kwargs)


def list_shares(client=None, **kwargs):
    return manila_client(client).shares.list(return_raw=True, **kwargs)


def delete_share(share_id, client=None, **kwargs):
    try:
        manila_client(client).shares.delete(share_id, **kwargs)
    except exceptions.NotFound:
        LOG.debug(f'Share {share_id} was not found')
        return False
    else:
        LOG.debug(f'Share {share_id} was deleted successfully')
        return True


def extend_share(share_id, new_size, client=None):
    return manila_client(client).shares.extend(share_id, new_size)


def get_share(share_id, client=None):
    try:
        return manila_client(client).shares.get(share_id, return_raw=True)
    except exceptions.NotFound as ex:
        raise _exceptions.ShareNotFound(id=share_id) from ex


def get_shares_by_name(share_name, client=None):
    share_list = list_shares(client=client)
    shares = [
        s for s in share_list if s['name'] == share_name
    ]
    return shares


def list_share_types(client=None):
    return manila_client(client).share_types.list()


def create_share_type(name, spec_driver_handles_share_servers,
                      client=None):
    return manila_client(client).share_types.create(
        name, spec_driver_handles_share_servers)


def ensure_default_share_type_exists(client=None):
    name = CONF.tobiko.manila.default_share_type_name
    dhss = CONF.tobiko.manila.spec_driver_handles_share_servers
    for share_type in list_share_types(client=client):
        if share_type.name == name:
            return
    create_share_type(name, dhss, client=client)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import pytest
from oslo_log import log
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova as nova_osp
from tobiko.openstack import octavia
from tobiko.openstack import topology
from tobiko.openstack import tests
from tobiko.tests.faults.ha import cloud_disruptions
from tobiko.tripleo import pacemaker
from tobiko.tripleo import processes
from tobiko.tripleo import containers
from tobiko.tripleo import nova
from tobiko.tripleo import overcloud
from tobiko.tripleo import undercloud


CONF = config.CONF
LOG = log.getLogger(__name__)
SKIP_MESSAGE_EXTLB = ('Tests requiring a main VIP should be skipped when an '
                      'external load balancer is used')
has_external_lb = CONF.tobiko.rhosp.has_external_load_balancer


def overcloud_health_checks(passive_checks_only=False,
                            skip_mac_table_size_test=False):
    # this method will be changed in future commit
    check_pacemaker_resources_health()
    check_overcloud_processes_health()
    nova.check_nova_services_health()
    tests.test_alive_agents_are_consistent_along_time()
    if not passive_checks_only:
        # create a uniq stack
        check_vm_create()
        nova.start_all_instances()
        nova.check_computes_vms_running_via_virsh()
    containers.list_node_containers.cache_clear()
    containers.assert_all_tripleo_containers_running()
    containers.assert_equal_containers_state()
    containers.run_container_config_validations()
    tests.test_ovn_dbs_validations()
    # skip_mac_table_size_test has to be removed when BZ1695122 is resolved
    # we need it for the moment because this validation should not be performed
    # after any overcloud node is rebooted
    if not skip_mac_table_size_test:
        tests.test_ovs_bridges_mac_table_size()

    octavia_health_checks()


# check vm create with ssh and ping checks
def check_vm_create():
    tests.test_server_creation()
    if overcloud.is_ovn_bgp_agent_running():
        try:
            node = topology.find_openstack_node(group='networker')
        except topology.NoSuchOpenStackTopologyNodeGroup:
            node = topology.find_openstack_node(group='controller')
        expose_tenant_networks = topology.get_config_setting(
            'bgp-agent.conf', node.ssh_client, 'expose_tenant_networks')
        if expose_tenant_networks and expose_tenant_networks.lower() == 'true':
            tests.test_server_creation_no_fip()


# check cluster failed statuses
def check_pacemaker_resources_health():
    return pacemaker.PacemakerResourcesStatus().all_healthy


def check_overcloud_processes_health():
    procs = processes.OvercloudProcessesStatus()
    return (procs.basic_overcloud_processes_running and
            procs.ovn_overcloud_processes_validations)


def octavia_health_checks():
    if keystone.is_service_missing(name='octavia'):
        LOG.debug('octavia service not available')
        return
    # check LBs and Amphorae are healthy
    LOG.debug("check all LBs are in healthy status")
    for lb in octavia.list_load_balancers():
        LOG.debug("checkin LBs:\n%s", lb)
        octavia.wait_for_status(lb['id'], interval=5., timeout=120.)
        octavia.wait_for_status(lb['id'],
                                status_key=octavia.OPERATING_STATUS,
                                status=octavia.ONLINE,
                                interval=5.,
                                timeout=120.)
    LOG.debug("All LBs are in healthy status")


class OvercloudHealthCheck(tobiko.SharedFixture):

    skips: typing.FrozenSet[str] = frozenset()

    @classmethod
    def run_before(cls, **params):
        cls.run(after=False, **params)

    @classmethod
    def run_after(cls, **params):
        cls.run(after=True, **params)

    @classmethod
    def run(cls, after: bool, **params):
        fixture = tobiko.get_fixture(cls)
        params.setdefault('passive_checks_only', False)
        # In version OSP17.0 and highier,
        # 'test_ovs_bridges_mac_table_size()' test can run.
        if topology.verify_osp_version('17.0', lower=True):
            params.setdefault('skip_mac_table_size_test', True)
        skips = frozenset(k for k, v in params.items() if v)
        if after or skips < fixture.skips:
            # Force re-check
            tobiko.cleanup_fixture(fixture)
        else:
            LOG.info("Will skip Overcloud health checks if already "
                     f"executed: {params}")
        fixture.skips = skips
        tobiko.setup_fixture(fixture)

    def setup_fixture(self):
        # run validations
        params = {name: True
                  for name in self.skips}
        LOG.info(f"Start executing Overcloud health checks: {params}.")
        overcloud_health_checks(**params)
        LOG.info(f"Overcloud health checks successfully executed: {params}.")

    def cleanup_fixture(self):
        self.skips = frozenset()


@undercloud.skip_if_missing_undercloud
class DisruptTripleoNodesTest(testtools.TestCase):
    """ HA Tests: run health check -> disruptive action -> health check
    disruptive_action: a function that runs some
    disruptive scenario on a overcloud"""
    vms_detailed_info = None
    undisrupt_network_at_teardown = False

    def tearDown(self):
        super(DisruptTripleoNodesTest, self).tearDown()

        # this loop is run after test_z999_reboot_controller_galera_main_vip
        for i, vm in enumerate(self.vms_detailed_info or []):
            if vm is None or vm.get('id') is None:
                # server_name pattern comes from multi_ip_test_stack.yaml
                server_name = "group_of_vms_%d" % i
                try:
                    server = nova_osp.find_server(unique=True,
                                                  name=server_name)
                except tobiko.ObjectNotFound:
                    LOG.debug(f"Server {server_name} not found. "
                              "Perhaps it was never created.")
                    continue
                vm_id = server.id
            else:
                vm_id = vm['id']
            try:
                nova_osp.delete_server(vm_id)
            except nova_osp.ServerNotFoundError:
                LOG.debug(f"Server {vm_id} not found. "
                          "Perhaps it was already deleted.")

        # this is run after test_network_disruptor_main_vip
        if self.undisrupt_network_at_teardown:
            self.undisrupt_network_at_teardown = False
            cloud_disruptions.network_undisrupt_controller_main_vip()

    def test_0vercloud_health_check(self):
        OvercloudHealthCheck.run_before(skip_mac_table_size_test=False)

    @nova.skip_background_vm_ping_checks_when_nondvr
    def test_z99_hard_reboot_controllers_recovery(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_all_controller_nodes()
        OvercloudHealthCheck.run_after()

    @staticmethod
    def _any_amphora_lb():
        """ returns True if octavia is configured and any octavia amphora LB
        exists"""
        return (keystone.has_service(name='octavia') and
                any([lb['provider'] == octavia.AMPHORA_PROVIDER
                     for lb in octavia.list_load_balancers()]))

    @nova.skip_background_vm_ping_checks
    def test_soft_reboot_computes_recovery(self):
        OvercloudHealthCheck.run_before()

        sequentially = self._any_amphora_lb()
        cloud_disruptions.reset_all_compute_nodes(hard_reset=False,
                                                  sequentially=sequentially)
        # verify VM status is updated after reboot
        nova.wait_for_all_instances_status('SHUTOFF')
        # start all VM instance
        # otherwise sidecar containers will not run after computes reboot
        nova.start_all_instances()
        OvercloudHealthCheck.run_after(passive_checks_only=True)

    # TODO(eolivare): the following test is skipped due to rhbz#1890895
    # def test_hard_reboot_computes_recovery(self):
    #     OvercloudHealthCheck.run_before()
    #     cloud_disruptions.reset_all_compute_nodes(hard_reset=True)
    #     # verify VM status is updated after reboot
    #     nova.wait_for_all_instances_status('SHUTOFF')
    #     # start all VM instance
    #     # otherwise sidecar containers will not run after computes reboot
    #     nova.start_all_instances()
    #     OvercloudHealthCheck.run_after(passive_checks_only=True)

    @nova.skip_background_vm_ping_checks_when_nondvr
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_z999_reboot_controller_galera_main_vip(self):
        # This test case may fail at times if RHBZ#2124877 is not resolved
        # but that bug is due to a race condition,
        # so it is not reproducible 100% times
        OvercloudHealthCheck.run_before(passive_checks_only=True)
        multi_ip_test_fixture, ports_before_stack_creation = \
            cloud_disruptions.reboot_controller_galera_main_vip()
        OvercloudHealthCheck.run_after(passive_checks_only=True)
        self.vms_detailed_info = cloud_disruptions.get_vms_detailed_info(
            multi_ip_test_fixture)
        LOG.debug('detailed info from the list of vms created: %r',
                  self.vms_detailed_info)
        cloud_disruptions.check_no_duplicate_ips(
            self.vms_detailed_info, ports_before_stack_creation)

    @nova.skip_background_vm_ping_checks_when_nondvr
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_z99_reboot_controller_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_controller_main_vip()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks_when_nondvr
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_z99_reboot_controller_non_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_controllers_non_main_vip()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks_when_nondvr
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_z99_crash_controller_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.crash_controller_main_vip()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks_when_nondvr
    @overcloud.skip_unless_kexec_tools_installed
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_z99_crash_controller_non_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.crash_controllers_non_main_vip()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks_when_nondvr
    @pacemaker.skip_if_fencing_not_deployed
    @testtools.skipIf(has_external_lb, SKIP_MESSAGE_EXTLB)
    def test_network_disruptor_main_vip(self):
        OvercloudHealthCheck.run_before()
        self.undisrupt_network_at_teardown = True
        cloud_disruptions.network_disrupt_controller_main_vip()
        OvercloudHealthCheck.run_after()

    # @pacemaker.skip_if_fencing_not_deployed
    # def test_network_disruptor_non_main_vip(self):
    #     OvercloudHealthCheck.run_before()
    #     cloud_disruptions.network_disrupt_controllers_non_main_vip()
    #     OvercloudHealthCheck.run_after()
    #     cloud_disruptions.network_undisrupt_controllers_non_main_vip()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_pcs_master_resource(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_pcs_master_resource()
        OvercloudHealthCheck.run_after()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_pcs_resource(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_pcs_resource()
        OvercloudHealthCheck.run_after()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_master_container(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_master_container()
        OvercloudHealthCheck.run_after()

    def test_kill_rabbitmq_service_one_controller(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.kill_rabbitmq_service()
        OvercloudHealthCheck.run_after()

    def test_kill_all_galera_services(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.kill_all_galera_services()
        OvercloudHealthCheck.run_after()

    def test_remove_all_grastate_galera(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.remove_all_grastate_galera()
        OvercloudHealthCheck.run_after()

    def test_remove_one_grastate_galera(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.remove_one_grastate_galera()
        OvercloudHealthCheck.run_after()

    def test_request_galera_sst(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.request_galera_sst()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks_when_nondvr
    @pytest.mark.flaky(reruns=0)
    def test_controllers_shutdown(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.test_controllers_shutdown()
        OvercloudHealthCheck.run_after()

    @overcloud.skip_unless_ovn_bgp_agent
    def test_restart_ovn_bgp_agents(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.restart_service_on_all_nodes(
            topology.get_agent_service_name(neutron.OVN_BGP_AGENT))
        OvercloudHealthCheck.run_after()

    @overcloud.skip_unless_ovn_bgp_agent
    def test_restart_frr(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.restart_service_on_all_nodes(
            topology.get_agent_service_name(neutron.FRR))
        OvercloudHealthCheck.run_after()

    def test_restart_neutron(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.restart_service_on_all_nodes(
            topology.get_agent_service_name(neutron.NEUTRON))
        OvercloudHealthCheck.run_after()

# [..]
# more tests to follow
# run health checks
# faults stop rabbitmq service on one controller
# run health checks again



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools

from oslo_log import log

import tobiko
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _hostname
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class CommandNotFound(tobiko.ObjectNotFound):
    message = "Command {command!r} not found on host {hostname!r}"


class SkipOnCommandNotFound(CommandNotFound, tobiko.SkipException):
    pass


@functools.lru_cache()
def find_command(command: str,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False,
                 skip=False) -> str:
    hostname = _hostname.get_hostname(ssh_client=ssh_client)
    try:
        result = _execute.execute(['which', command],
                                  ssh_client=ssh_client,
                                  sudo=sudo)
    except _exception.ShellCommandFailed as ex:
        if skip:
            raise SkipOnCommandNotFound(command=command,
                                        hostname=hostname) from ex
        else:
            raise CommandNotFound(command=command,
                                  hostname=hostname) from ex
    else:
        command_path = result.stdout.strip()
        LOG.debug(f"Command {command!r} found on host {hostname}: "
                  f"{command_path}")
        return command_path



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

from tobiko.openstack import neutron
from tobiko.openstack import tests
from tobiko.tripleo import overcloud
from tobiko.tripleo import undercloud


@neutron.skip_unless_is_ovn()
@undercloud.skip_if_missing_undercloud
@overcloud.skip_unless_ovn_using_raft
class TestRAFTDisruption(testtools.TestCase):

    def test_raft_status(self):
        tests.test_raft_cluster()
        tests.test_raft_clients_connected()

    def test_ovsdb_transation(self):
        tests.test_ovsdb_transactions()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


def get_file_size(file_name: str,
                  ssh_client: ssh.SSHClientType = None,
                  sudo: bool = None) \
        -> int:
    output = sh.execute(f'wc -c "{file_name}"',
                        ssh_client=ssh_client,
                        sudo=sudo).stdout.strip()
    size, _ = output.split(' ', 1)
    return int(size)


def assert_file_size(file_size: int,
                     file_name: str,
                     ssh_client: ssh.SSHClientType = None,
                     sudo: bool = None):
    size = get_file_size(file_name=file_name,
                         ssh_client=ssh_client,
                         sudo=sudo)
    tobiko.get_test_case().assertEqual(file_size, size)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing

from packaging import version as _version

from tobiko import _exception


VERSION_PATTERN = re.compile(r"[0-9]*\.[0-9]*\.[0-9]*")


class InvalidVersion(_exception.TobikoException):
    message = "invalid version: {text}"


class VersionMismatch(_exception.TobikoException):
    message = "version mismatch {version} {cause}"
    cause = ''


try:
    VERSION_CLASSES = _version.LegacyVersion, _version.Version
except AttributeError:
    VERSION_CLASSES = _version.Version

Version = typing.Union[_version.Version]
VersionType = typing.Union[Version, str]


def get_version(obj: VersionType):
    if isinstance(obj, VERSION_CLASSES):
        return obj
    elif isinstance(obj, str):
        return parse_version(obj)
    else:
        raise TypeError(f'Cannot get version from object {obj!r}')


def parse_version(text: str) -> VersionType:
    match = VERSION_PATTERN.search(text.strip())
    if match is not None:
        text = match.group()
    try:
        return _version.Version(text)
    except _version.InvalidVersion as ex:
        raise InvalidVersion(text=text) from ex


def match_version(actual: VersionType,
                  min_version: VersionType = None,
                  max_version: VersionType = None) -> bool:
    try:
        check_version(actual=actual,
                      min_version=min_version,
                      max_version=max_version)
    except VersionMismatch:
        return False
    else:
        return True


def check_version(actual: VersionType,
                  min_version: VersionType = None,
                  max_version: VersionType = None,
                  mismatch_error=VersionMismatch):
    actual = get_version(actual)
    if min_version is not None:
        min_version = get_version(min_version)
        if actual < min_version:
            raise mismatch_error(
                version=actual,
                cause=f">= {max_version}")

    if max_version is not None:
        max_version = get_version(max_version)
        if actual >= max_version:
            raise mismatch_error(
                version=actual,
                cause=f">= {max_version}")
    return True



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect

from oslo_log import log
import testtools

from tobiko.common import _exception
from tobiko.common import _fixture
from tobiko.common import _skip


LOG = log.getLogger(__name__)


def runs_operation(operation_object):
    return RunsOperationProperty(operation_object)


class RunsOperations(object):

    def setUp(self):
        super(RunsOperations, self).setUp()  # pylint: disable=no-member
        for operation_name in get_runs_operation_names(self):
            get_operation(operation_name)


def add_runs_operation_name(obj, name):
    required_names = getattr(obj, '__tobiko_runs_operations__', None)
    if not isinstance(required_names, set):
        if required_names:
            required_names = set(required_names)
        else:
            required_names = set()
        # cache list for later use
        obj.__tobiko_runs_operations__ = required_names
    required_names.add(name)


def get_runs_operation_names(obj):
    '''Get operation names required by given :param obj:'''

    name, obj = _fixture.get_name_and_object(obj)

    operations = getattr(obj, '__tobiko_runs_operations__', set())
    if isinstance(operations, tuple):
        return operations

    try:
        # froze required names to avoid re-discovering them again
        obj.__tobiko_runs_operations__ = tuple()
    except AttributeError:
        # ignore this object because not defined by Python code
        return tuple()

    LOG.debug('Discover operations for object %r', name)
    if operations:
        operations = set(operations)

    if inspect.ismethod(obj) or inspect.isfunction(obj):
        pass  # decorated functions have __tobiko_runs_operations__ defined

    else:
        if not inspect.isclass(obj):
            obj = type(obj)
        operations.update(get_runs_operation_names_from_class(obj))

    # Return every operation in alphabetical order and
    # froze required names to avoid re-discovering them again
    operations = tuple(sorted(operations))
    obj.__tobiko_runs_operations__ = operations
    return operations


def get_runs_operation_names_from_class(cls):
    """Get list of members of type RequiredFixtureProperty of given class"""

    # inspect.getmembers() would iterate over such many testtools.TestCase
    # members too, so let exclude members from those very common base classes
    # that we know doesn't have members of type RunsOperationProperty
    base_classes = cls.__mro__
    for base_class in [testtools.TestCase, _fixture.SharedFixture]:
        if issubclass(cls, base_class):
            base_classes = base_classes[:base_classes.index(base_class)]
            break

    # Get all members for selected class without calling properties or methods
    members = {}
    for base_class in reversed(base_classes):
        members.update(base_class.__dict__)

    # Return all member operation names
    operations = set()
    for name, member in sorted(members.items()):
        if not name.startswith('__'):
            operations.update(get_runs_operation_names(obj=member))
    return operations


class RunsOperationProperty(object):

    def __init__(self, obj):
        self.name, self.obj = _fixture.get_name_and_object(obj)

    @property
    def __tobiko_runs_operations__(self):
        return (self.name,)

    def before_operation(self, obj):

        def is_before():
            return self.get_operation().is_before

        decorator = _skip.skip_unless(f"Before operation {self.name}",
                                      is_before)

        return self.with_operation(decorator(obj))

    def after_operation(self, obj):

        def is_after():
            return self.get_operation().is_after

        decorator = _skip.skip_unless(f"After operation {self.name}",
                                      is_after)

        return self.with_operation(decorator(obj))

    def with_operation(self, obj):
        add_runs_operation_name(obj, self.name)
        return obj

    def __get__(self, instance, _):
        if instance is None:
            return self
        else:
            return self.get_operation()

    def get_operation(self):
        return get_operation(self.obj)


def before_operation(obj):
    return RunsOperationProperty(obj).before_operation


def after_operation(obj):
    return RunsOperationProperty(obj).after_operation


def with_operation(obj):
    return RunsOperationProperty(obj).with_operation


def get_operation(obj):
    operation = _fixture.get_fixture(obj)
    _exception.check_valid_type(operation, Operation)

    if operation.is_before and operation_config().run_operations:
        operation = _fixture.setup_fixture(obj)

    return operation


def get_operation_name(obj):
    return _fixture.get_object_name(obj)


def operation_config():
    return _fixture.setup_fixture(OperationsConfigFixture)


class OperationsConfigFixture(_fixture.SharedFixture):

    after_operations = None
    run_operations = None

    def setup_fixture(self):
        from tobiko import config
        self.after_operations = set(
            config.get_list_env('TOBIKO_AFTER_OPERATIONS'))
        self.run_operations = config.get_bool_env(
            'TOBIKO_RUN_OPERATIONS') or False


class Operation(_fixture.SharedFixture):

    @property
    def operation_name(self):
        return _fixture.get_fixture_name(self)

    @property
    def is_before(self):
        return not self.is_after

    _is_after = None

    @property
    def is_after(self):
        is_after = self._is_after
        if is_after is None:
            config = operation_config()
            is_after = self.operation_name in config.after_operations
            self._is_after = is_after
        return is_after

    def setup_fixture(self):
        if not self.is_after:
            LOG.debug('Executing operation: %r', self.operation_name)
            try:
                self.run_operation()
            except Exception as ex:
                LOG.debug('Operation %r failed: %r', self.operation_name, ex)
                raise
            else:
                LOG.debug('Operation executed: %r', self.operation_name)
            finally:
                self._is_after = True

    def run_operation(self):
        raise NotImplementedError
